# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Iterator and functional programming tools. 
Most of the iterator methods were copied from the 
Python itertools documentation with added docstrings
to show example usage.

"""

import operator
from operator import getitem, setitem, delitem
from itertools import *

try:
    from operator import attrgetter
    
except ImportError:
    def attrgetter(attr):
        """ Returns a function f such that
        calling f(ob) returns ob.attr:
        
            >>> class Test: pass
            >>> test = Test()
            >>> test.aProp = "Test Value"
            
            >>> fn = attrgetter('aProp')
            >>> fn(test)
            'Test Value'
            
        If ob.attr is not found the normal attribute
        error is raised:
            
            >>> fn = attrgetter('oops')
            >>> print fn(test)
            Traceback (most recent call last):
                ...
            AttributeError: Test instance has no attribute 'oops'

        
        """
        return lambda obj: getattr(obj, attr)
        
def attrsetter(attr):
    """ Returns a function f such that
    calling f(ob,value) sets ob.attr = value:
    
            >>> class Test: pass
            >>> test = Test()
            >>> fn = attrsetter('aProp')
            >>> fn(test,"Test Value")
            >>> test.aProp 
            'Test Value'
            
            >>> fn(test,"Changed Value")
            >>> test.aProp
            'Changed Value'
    
    """

    return lambda obj, value: setattr(obj, attr, value)
    
def attrdeleter(attr):
    """ Returns a function f such that
    calling f(ob) deletes ob.attr:
    
        >>> class Test: pass
        >>> test = Test()
        >>> test.aProp = "Test Value"
        >>> test.aProp
        'Test Value'
        
        >>> fn = attrdeleter('aProp')
        >>> fn(test)
        >>> print test.aProp
        Traceback (most recent call last):
            ...
        AttributeError: Test instance has no 
            attribute 'test'
        
    """
    
    return lambda obj: delattr(obj, attr)

try:
    from operator import itemgetter
    
except ImportError:
    def itemgetter(item):
        """ Returns a function f such that
        calling f(ob) returns ob[item]:
        
            >>> test = { 'item' : 'Test Value' }
            >>> test['item']
            'Test Value' 
            
            >>> fn = itemgetter('item')
            >>> fn(test)
            'Test Value' 
            
        If ob[item] is not found the normal key
        error is raised:
            
            >>> fn = itemgetter('oops')
            >>> print fn(test)
            Traceback (most recent call last):
                ...
            KeyError: 'oops'
        
        """
        return lambda obj: getitem(obj, item)

def itemsetter(item):
    """ Returns a function f such that
    calling f(ob,key) sets ob[attr] = value:
    
        >>> test = {}
        >>> fn = itemsetter('item')
        
        >>> fn(test, 'Test Value')
        >>> test['item']
        'Test Value'
        
        >>> fn(test, 'Changed Value')
        >>> test['item']
        'Changed Value'
        
    """
    return lambda obj, value: setitem(obj, item, value)
    
def itemdeleter(item):
    """ Returns a function f such that
    calling f(ob) deletes f[item]:
    
        >>> test = {'item' : 'Test Value'}
        >>> test['item']
        'Test Value'
        
        >>> fn = itemdeleter('item')
        >>> fn(test)
        
        >>> test['item']
        Traceback (most recent call last):
            ...
        KeyError: 'item'
        
    """
    return lambda obj: delitem(obj, item)


def exhaust(seq):
    """ Iterate thorugh all items in seq discarding
    return values.
    
        >>> test = xrange(10)
        >>> print exhaust(test)
        None
    
    """
    for i in seq:
        pass

def first(seq):
    """ 
    Return the first item in seq and discard the rest:
        
        >>> first([1, 2, 3])
        1
        
        >>> print first([])
        Traceback (most recent call last):
            ...
        StopIteration


    """
    for first in seq:
        return first
    else:
        raise StopIteration()

        
def take(n, seq):
    """ 
    Return the first n elements of seq:
        
        >>> take(3, [1, 2, 3, 4])
        [1, 2, 3]
        
        >>> take(3, [])
        []
    
    """
    return list(islice(seq, n))

def ienumerate(iterable):
    """ Returns a iterator that yields
    (0, item1), (1, item2), ...:
        
        >>> list(ienumerate([1, 2, 3]))
        [(0, 1), (1, 2), (2, 3)]

        >>> list(ienumerate([]))
        []
        
    """
    return izip(count(), iterable)

def tabulate(function):
    """ 
    Return function(0), function(1), ...:
    
        >>> iter = tabulate(lambda n: n*2)
        >>> [iter.next() for n in xrange(10)]
        [0, 2, 4, 6, 8, 10, 12, 14, 16, 18]

    """
    return imap(function, count())

def iteritems(mapping):
    """ Returns a iterator yielding (key, value)
    pairs from mapping:
        
        >>> i = iteritems({ 'a' : 0, 'b' : 1, 'c' : 2})
        >>> i = list(i)
        >>> i.sort()
        >>> i
        [('a', 0), ('b', 1), ('c', 2)]
    
    """
    return izip(mapping.iterkeys(), mapping.itervalues())

def nth(iterable, n):
    """ 
    Returns the nth item of iterable. 
    
        >>> nth([1, 2, 3], 1)
        [2]
        
    """
    return list(islice(iterable, n, n+1))

def all(seq, pred=bool):
    """ Returns True if pred(x) is True for every 
    element in the iterable.
    
        >>> all([True, True, True])
        True
    
        >>> all([True, True, False])
        False

        >>> all([])
        True

    """
    for elem in ifilterfalse(pred, seq):
        return False
    return True

def any(seq, pred=bool):
    """ Returns True if pred(x) is True for at least one 
    element in the iterable:
        
        >>> any([False, True, False])
        True
    
        >>> any([False, False, False])
        False

        >>> any([])
        False
        
    
    """
    for elem in ifilter(pred, seq):
        return True
    return False

def no(seq, pred=bool):
    """ Returns True if pred(x) is False for every 
    element in the iterable.
    
        >>> no([False, False, False])
        True
        
        >>> no([True, False, False])
        False
        
        >>> no([])
        True
    """
    for elem in ifilter(pred, seq):
        return False
    return True

def quantify(seq, pred=bool):
    """ Count how many times the predicate is True 
    in the sequence.
    
        >>> quantify([True, False, True])
        2
        
    """
    return sum(imap(pred, seq))

def pad(seq, padvalue = None):
    """ Returns the sequence elements and then returns None 
    indefinitely. Useful for emulating the behavior of the 
    built-in map() function.
    
        >>> i = pad([1, 2, 3]).next
        >>> [i() for n in xrange(5)]
        [1, 2, 3, None, None]
        >>> i = pad([1, 2, 3], "Pad").next
        >>> [i() for n in xrange(5)]
        [1, 2, 3, 'Pad', 'Pad']
        
    """
    
    return chain(seq, repeat(padvalue))

def ncycles(seq, n):
    """ 
    Returns the sequence elements n times:
    
        >>> list(ncycles([1, 2, 3], 3))
        [1, 2, 3, 1, 2, 3, 1, 2, 3]
        
        >>> list(ncycles([], 3))
        []
        
        >>> list(ncycles([1, 2, 3], 0))
        []

    """
    return chain(*repeat(seq, n))

def dotproduct(vec1, vec2):
    """ 
    Calculate the dot product of two vectors:
    
        >>> dotproduct([1, 2, 3], [4, 5, 6])
        32
        
    """
    
    return sum(imap(operator.mul, vec1, vec2))

def flatten(*seqs):
    """ 
    Create a single list from seqs:
        
        >>> flatten([1, 2, 3], (4, 5, 6))
        [1, 2, 3, 4, 5, 6]
    
    """
    return list(chain(*seqs))

def repeatfunc(func, times=None, *args):
    """ 
    Repeat calls to func with specified arguments: 
    
        >>> i = repeatfunc(lambda: 1)
        >>> [i.next() for n in xrange(10)]
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]

        >>> list(repeatfunc(lambda: 2, 10))
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]

    """
    
    if times is None:
        return starmap(func, repeat(args))
    else:
        return starmap(func, repeat(args, times))

def pairwise(iterable):
    """ 
    Group iterable into pairs:
    
        >>> list(pairwise([1, 2, 3, 4]))
        [(1, 2), (2, 3), (3, 4)]

    """
    
    a, b = tee(iterable)
    try:
        b.next()
    except StopIteration:
        pass
    return izip(a, b)

def grouper(n, iterable, padvalue=None):
    """ Groups iterable into tuples of n elements:
    
        >>> list(grouper(3, 'abcdefg', 'x'))
        [('a', 'b', 'c'), ('d', 'e', 'f'), ('g', 'x', 'x')]

    """
    
    return izip(*[chain(iterable, repeat(padvalue, n-1))]*n)

def forever(iter):
    """ Generator that loops forever over a sequence.
    
        >>> test = forever([1,2,3])
        >>> [test.next() for i in range(10)]
        [1, 2, 3, 1, 2, 3, 1, 2, 3, 1]

    """
    while True:
        for item in iter:
            yield item


class Forever(object):
    """ Iterator that loops forever over a sequence.
    
        >>> queue = Forever([1, 2, 3])
        >>> [queue.next() for i in range(10)]
        [1, 2, 3, 1, 2, 3, 1, 2, 3, 1]

    """


    def __init__(self, items):
        self._iter = forever(items)
        
    def __iter__(self):
        return self
        
    def next(self):
        return self._iter.next()



def testSuite():
    """ Returns doctest testsuite for module. """
    
    import doctest as dt
    return dt.DocTestSuite(optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
                dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)

if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())

    
    
