# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Utility methods and classes. """

import sys, types
from rstdoc.misclibs.functional import count, any

__all__ = ["UNDEFINED", "Enum", "Once", "Property", "Constant", 
    "DefaultProperty", "abstractMethod", "initAttrs", 
    "toUnicode", "toString"]

class UNDEFINED(object):
    """ A unique value for keyword argument defaults. """
    
UNDEFINED = UNDEFINED()

def toUnicode(s, encoding = "iso-8859-1", errors = "replace"):
    if isinstance(s, types.UnicodeType):
        return s
    elif isinstance(s, types.StringType):
        return s.decode(encoding, errors)
    else:
        return unicode(s)

def toString(s, encoding = "iso-8859-1", errors = "replace"):
    if isinstance(s, types.StringType):
        return s
    elif isinstance(s, types.UnicodeType):
        return s.encode(encoding, errors)
    else:
        return str(s)

def printable(s):
    s = toString(s, "utf-8")
    return s.encode("string_escape").replace("\\'","'")

def initAttrs(instance, **attrs):
    for k,v in attrs.items():
        setattr(instance, k, v)

class Enum(object):
    """ Defines an object with attributes that enumerate 
    a sequence of values:
        
        >>> class SomeClass(object):
        ...     enum = Enum("zero", "one", "two")
        >>> SomeClass().enum.zero
        0
        >>> SomeClass().enum.one
        1
        >>> SomeClass().enum.two
        2
        >>> SomeClass().enum.oops
        Traceback (most recent call last):
            ...
        AttributeError: Invalid attribute for Enum: 'oops'

    The starting value can also be set:
        
        >>> class SomeClass(object):
        ...     enum = Enum("zero", "one", "two", 
        ...         start = 1000)
        >>> SomeClass().enum.zero
        1000
        >>> SomeClass().enum.one
        1001
        >>> SomeClass().enum.two
        1002

    The Enum class can also be used outside a class definition:
        
        >>> test = Enum("zero", "one", "two")
        >>> test.zero
        0
        
    Enum values can be iterated over as follows (producing
    a series of (attr, value) pairs):
        
        >>> [v for v in test]
        [('zero', 0), ('one', 1), ('two', 2)]

    The index operator will return the string value of the Enum:
        
        >>> test[test.zero]
        'zero'

    """
    
    def __init__(self, *values, **kw):
        self.values = list(values)
        self.start = kw.get("start", 0)

    def __iter__(self):
        for value in self.values:
            yield value, getattr(self, value)
            
    def __getattr__(self, attr):
        try:
            return self.values.index(attr) + self.start
        except ValueError:
            raise AttributeError(
                "Invalid attribute for Enum: '%s'" % (attr,))
    
    def __getitem__(self, item):
        return self.values[item]
    
class PropertyClass(type):
    """ Meta class for Property class. """
    
    nextId = count().next
    def __new__(klass, name, bases, cdict):
        cdict["__name__"] = name
        #cdict["__varname__"] = "__%s_%05i" % (name, klass.nextId())
        cls = type.__new__(klass, name, bases, cdict)
        
        # If the class atrribute 'isClass' is True
        # then a new class is being defined.
        if 'isClass' in cdict:
            del cls.isClass # Do not propagate to subclasses
            return cls
            
        # Otherwise create an instance 
        else:
            return cls()
 
class Property(object):
    """ Creates a descripter class (property). 
    
    To create a attribute with default fget, fset, and 
    fdel methods simply subclass Property:
        
        >>> import inspect
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         "My Property"

        >>> test = SomeClass()
        >>> test.aProp = 1
        >>> test.aProp
        1
    
        >>> inspect.getdoc(SomeClass.aProp)
        'My Property'

    It is also possible to assign a default value:
    
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         default = "Default Value"
        
        >>> test = SomeClass()
        >>> test.aProp
        'Default Value'
        
        >>> test.aProp = "Assigned Value"
        >>> test.aProp
        'Assigned Value'

    If the attribute  is deleted then the attribute
    reverts to the default value (if any):
        
        >>> del test.aProp
        >>> test.aProp
        'Default Value'

    The default can also be defined as a 2-argument
    callable:

        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         fset = fget = fdel = True
        ...         def default(self, instance):
        ...             return instance
        >>> SomeClass.aProp
        <...util.aProp object at 0x...>

    If no default value is given, and no assignment to
    the attribute is perfomed, then an AttributeError
    is raised:
        
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         "My Property"
        ...         fset = fget = fdel = True
        
        >>> SomeClass().aProp
        Traceback (most recent call last):
            ...
        AttributeError: Undefined attribute 'aProp' 
                with no default given

    Custom fget, fset, and fdel methods can also be
    provided:
        
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         default = "Default Value"
        ...         fset = fget = fdel = True
        ...         def fget(self, instance):
        ...             return ('custom: ' + 
        ...                 self.getDefault(instance))
        >>> SomeClass().aProp
        'custom: Default Value'
        
    When defining a custom method, the getValue, getDefault,
    and setValue methods can be used to assign values to
    a unique attribute of the 'instance'.  The method 
    argument 'self' refers to the Property instance,  the
    argument 'instance' refers to the containing class:
        
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         fset = fget = fdel = True
        ...         def fget(self, instance):
        ...             print "Self:", self
        ...             print "Instance:", instance
        >>> SomeClass().aProp
        Self: <...util.aProp object at 0x...>
        Instance: <...util.SomeClass object at 0x...>

    Therefore, values assigned to self will be shared among
    *all instances of the containing class:*
    
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         fset = fget = fdel = True
        ...         def fset(self, instance, value):
        ...             self.test = value
        ...         def fget(self, instance):
        ...             return self.test
        
        >>> test1 = SomeClass()
        >>> test2 = SomeClass()
        >>> test1.aProp = 1
        >>> test2.aProp = 2
        >>> test1.aProp
        2
        
    Each subclass of Property is assigned two special
    attributes which can be used to create unique
    variables in the instance's dictionary:
        
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         def fget(self, instance):
        ...             print '__name__ = ', self.__name__
        ...             print '__varname__ = ', self.__varname__
        >>> SomeClass().aProp
        __name__ =  aProp
        __varname__ =  __aProp_000...

    The Property.getValue, Property.getDefault, and 
    Property.setValue methods can be used to assign
    and retrieve values from the instance's dictionary.
    
    If fget, fset, fdel are assigned the value None then
    the corosponding attribute access is not possible:
        
        >>> class SomeClass(object):
        ...     class aProp(Property):
        ...         fget = fset = fdel = None
        >>> SomeClass().aProp
        Traceback (most recent call last):
            ...
        AttributeError: Can't get attribute

    """
    
    __metaclass__ = PropertyClass
    isClass = True
    default = UNDEFINED
    fget = fset = fdel = True
 
    def __init__(self):
        self.__varname__ = "__%s_%05i" % (
            self.__name__, 
            self.__class__.nextId())

    
    def getValue(self, instance):
        if not hasattr(instance, self.__varname__):
            return UNDEFINED
        else:
            return getattr(instance, self.__varname__)
        
    def setValue(self, instance, value):
        return setattr(instance, self.__varname__, value)
    
    def delValue(self, instance):
        if hasattr(instance, self.__varname__):
            delattr(instance, self.__varname__)
            
    def getDefault(self, instance):
        if callable(self.default):
            return self.default(instance)
        else:
            return self.default
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        elif self.fget is None:
            raise AttributeError("Can't get attribute")
        elif callable(self.fget):
            return self.fget(instance)
        elif self.getValue(instance) is not UNDEFINED:
            return self.getValue(instance)
        elif self.getDefault(instance) is not UNDEFINED:
            return self.getDefault(instance)
        else:
            raise AttributeError("Undefined attribute '%s' "
            "with no default given" % self.__name__)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("Can't set attribute")
        elif callable(self.fset):
            self.fset(instance, value)
        else:
            self.setValue(instance, value)
        
    def __delete__(self, instance):
        if self.fdel is None:
            raise AttributeError("Can't delete attribute")
        elif callable(self.fdel):
            self.fdel(self, instance)
        else:
            self.delValue(instance)
            
class Once(Property):
    """ Creates a property that is calculated at most
    1 time.  This idea was borrowed from the PEAK package.
    In the following example,  note that  the 'getValue' 
    method is only executed once:
        
        >>> import inspect
        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         "Some Doc String"
        ...
        ...         def value(self, instance):
        ...             print "This is executed once"
        ...             return "Some Value"
        
        >>> test = SomeClass()
        >>> test.aProp
        This is executed once
        'Some Value'
        >>> test.aProp
        'Some Value'
        
        >>> inspect.getdoc(SomeClass.aProp)
        'Some Doc String'
        
    If the attribute is deleted,  then the getValue method
    will be executed again:

        >>> del test.aProp
        >>> test.aProp
        This is executed once
        'Some Value'
        >>> test.aProp
        'Some Value'

    The value method or class attribute **must** be defined:

        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         pass
        Traceback (most recent call last):
            ...
        AttributeError: A class attribute or method 
            'value' must be declared for the Once 
            attribute 'aProp'


        
    The **Once** class is a subclass of **Property**, however, 
    the fget, fset, fdel, and default methods/attributes must 
    **not** be defined:
        
        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         def value(self, instance): pass
        ...         def fget(self, instance): pass
        Traceback (most recent call last):
            ...
        AttributeError: The Once attribute 'aProp' 
            cannot define fget/fset/fdel/default 
            methods or attributes


        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         def value(self, instance): pass
        ...         def fset(self, instance): pass
        Traceback (most recent call last):
            ...
        AttributeError: The Once attribute 'aProp' 
            cannot define fget/fset/fdel/default 
            methods or attributes

        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         def value(self, instance): pass
        ...         def fdel(self, instance): pass
        Traceback (most recent call last):
            ...
        AttributeError: The Once attribute 'aProp' 
            cannot define fget/fset/fdel/default 
            methods or attributes

        >>> class SomeClass(object):
        ...     class aProp(Once):
        ...         def value(self, instance): pass
        ...         def default(self, instance): pass
        Traceback (most recent call last):
            ...
        AttributeError: The Once attribute 'aProp' 
            cannot define fget/fset/fdel/default 
            methods or attributes

    Attribute assignment is not allowed:
        
        >>> test.aProp = "Oops"
        Traceback (most recent call last):
            ...
        AttributeError: Can't set attribute
        
    """
    isClass = True
    fget = fset = fdel = default = None
    
    def __init__(self):
        self.__varname__ = "__%s_%05i" % (
            self.__name__, 
            self.__class__.nextId())
            
        if not hasattr(self, "value"):
            raise AttributeError("A class attribute or "
                "method 'value' must be declared for "
                "the Once attribute '%s'" % self.__name__)
        if any([getattr(self, value, None) 
                for value in 
                    "fget", "fset", "fdel", "default"]):
            raise AttributeError("The Once attribute '%s' "
                "cannot define fget/fset/fdel/default "
                "methods or attributes" % self.__name__)
        
    def __get__(self, instance, owner):
        if instance is None:
            return self
            
        if self.getValue(instance) is UNDEFINED:
            if callable(self.value):
                self.setValue(instance, self.value(instance))
            else:
                self.setValue(instance, self.value)
        return self.getValue(instance)

    def __delete__(self, instance):
        if hasattr(instance, self.__varname__):
            delattr(instance, self.__varname__)
    
def Constant(value, doc = None):
    """ Returns a property that returns a constant
    value. 
    
    >>> import inspect
    >>> class SomeClass(object):
    ...     aProp = Constant(5, doc = "Some Doc String")
    >>> SomeClass().aProp
    5
    >>> inspect.getdoc(SomeClass.aProp)
    'Some Doc String'
    
    """
    
    return property(fget = lambda self: value, doc = doc)

 
def DefaultProperty(default = UNDEFINED):
    """ Creates a Property instance with an optional default
    value.  

    If a default is given, then it used as the initial value
    of the attribute:
        
        >>> class SomeClass(object):
        ...     aProp = DefaultProperty(5)
        >>> test = SomeClass()
        >>> test.aProp
        5
        
    The attribute can be set normally:
        
        >>> test.aProp = 1
        >>> test.aProp
        1
        
    If no default is given then the value of the property must 
    be set before it is accessed:
    
        >>> class SomeClass(object):
        ...     aProp = DefaultProperty()
        >>> test = SomeClass()
        >>> test.aProp
        Traceback (most recent call last):
            ...
        AttributeError: Undefined attribute 'Property' 
            with no default given
            
    Once the attribute has been set then it can be accessed 
    normally:
        
        >>> test.aProp = 1
        >>> test.aProp
        1
    
    """
    prop = Property()
    prop.fset = prop.fget = prop.fdel = True
    prop.default = default
    return prop
    
def abstractMethod(fn):
    """ The method decorated by the function will
    raise a NotImplementedError if called:
        
        >>> class Test:
        ...     @abstractMethod
        ...     def test(*args, **kw):
        ...         " Override me "
        ...         pass
        >>> Test().test()
        Traceback (most recent call last):
            ...
        NotImplementedError: Call of abstract method: 'test'
        
        >>> print Test().test.__doc__
        Override me
    """
    
    def abstractMethod(self, *args, **kw): 
        raise NotImplementedError(
            "Call of abstract method: '%s'" % fn.__name__)
            
    abstractMethod.__doc__ = fn.__doc__
    return abstractMethod

def testSuite():
    """ Returns doctest testsuite for module. """
    
    import util as moduleToTest
    import doctest as dt
    return dt.DocTestSuite(moduleToTest,
            optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
                dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)

if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())
