# -*- coding: iso-8859-1 -*-
#
# $Date: 2007-01-25 22:17:38 -0800 (Thu, 25 Jan 2007) $
# $Revision: 187 $
# $Author: schurch $
# $Id: __init__.py 187 2007-01-26 06:17:38Z schurch $
#
# Copyright Shawn L. Church (c) 2007, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" This module implements the Trac components for the 
RstDoc `TracPlugin`. """

import sys, os, os.path, urllib
from textwrap import dedent

from trac.core import *
from trac.web.chrome import INavigationContributor, \
    ITemplateProvider, add_stylesheet
from trac.web.main import IRequestHandler
from trac.util import TracError, escape, shorten_line
from trac.wiki.api import IWikiSyntaxProvider, \
    IWikiMacroProvider
from trac.Search import ISearchSource

from docutils import core, nodes, utils
from docutils.core import publish_doctree, \
        publish_from_doctree
from docutils.utils import new_document

from rstdoc.misclibs.util import *
from rstdoc.misclibs.functional import *
from rstdoc.rstlib import astdoc
from rstdoc.traclib import roles, wiki_filter
try:
    from trac.util import Markup
except ImportError:
    def Markup(markup): return markup
    
class RstDocRequestHandler(Component):
    """ I must be enabled if any other rstdoc components are
    enabled. """
    
    implements(IRequestHandler, ITemplateProvider)
        
    # IRequestHandler methods.
    def match_request(self, req):
        return req.path_info.startswith('/rstdoc')
    
    def process_request(self, req):
        documenter = Documenter(
            self.env, self.config)
            
        target = req.path_info[7:]
        documenter.processRequest(req, target)
        return "rstdoc.cs", None
        
    # ITemplateProvider methods
    def get_templates_dirs(self):
        from pkg_resources import resource_filename
        return [resource_filename(__name__, 'templates')]

    def get_htdocs_dirs(self):
        from pkg_resources import resource_filename
        return [('rstdoc', resource_filename(__name__, 'htdocs'))]

class RstDocNav(Component):
    """ If enabled, a `RstDoc` menu item will be added 
    to the Trac main menu. """
    
    implements(INavigationContributor)

    # INavigationContributor methods
    def get_active_navigation_item(self, req):
        return 'rstdoc'
                
    def get_navigation_items(self, req):
        yield 'mainnav', 'rstdoc', Markup('<a href="%s">RstDoc</a>' \
                                  % escape(self.env.href.rstdoc()))


class RstDocWikiSyntax(Component):
    """ If enabled, the `rstdoc:target` wiki syntax can be
    used to link to documentation."""

    implements(IWikiSyntaxProvider)

    # IWikiSyntaxProvider methods
    def get_wiki_syntax(self):
        return []

    def get_link_resolvers(self):
        yield ('rstdoc', self._rstdoc_formatter)

    def _rstdoc_formatter(self, formatter, ns, object, label):
        objects = urllib.unquote(object).split(":")
        object = objects and objects.pop(0)
        labels = urllib.unquote(label).split(':')
        label = labels and labels.pop()
        return '<a class="wiki" href="%s">%s</a>' % \
               (formatter.href.rstdoc(object), label)

class RstDocWikiMacro(Component):
    """ When enabled, generates documentation from the repository.
    
        Macro syntax: `[[rstdoc(target)]]`

    """

    implements(IWikiMacroProvider)

    # IWikiMacroProvider methods
    def get_macros(self):
        yield 'rstdoc'

    def get_macro_description(self, name):
        return self.__doc__
    
    def render_macro(self, req, name, content):
        target = content.split(" ", 1)[0]
        documenter = Documenter(self.env, self.config)
        return documenter.processRequest(req, target)

class RstFileWikiMacro(Component):
    """ When enabled, display reStructured text inline from
    a file in the repository.
    
        Macro syntax: `[[rstfile(target)]]`

    """

    implements(IWikiMacroProvider)

    # IWikiMacroProvider methods
    def get_macros(self):
        yield 'rstfile'

    def get_macro_description(self, name):
        return self.__doc__
    
    def render_macro(self, req, name, content):
        target = content.split(" ", 1)[0]
        documenter = Documenter(self.env, self.config)
        return documenter.processRequest(req, target, True)


class Documenter(object):
    """ I generate the documenation as required by the
    various components."""
    
    def __init__(self, env, config):
        self.env = env
        self.config = config
        self.root = self.config.get(
                'rstdoc', 'root', "/trunk")
        len = self.config.get(
                'rstdoc', 'doc.length', 1)
        try:
            self.docStringLength = int(len)
        except ValueError:
            self.docStringLength = 1

    def processRequest(self, req, target, rstFile = False):
        add_stylesheet(req, 'rstdoc/css/rstdoc.css')
        astdoc.setDocStringFilter(wiki_filter.wikiFilter)
        roles.setTracRef(
            roles.tracRef(
                self.env.href()))
        repo = self.env.get_repository()
        req.hdf['trac.href.rstdoc'] = self.env.href.rstdoc()
        if not target:
            target = self.root
        elif target.startswith('/'):
            pass
        else:
            target = self.root + "/" + target
        
        try:
            node =  repo.get_node(target)
            if node.get_content() is None:
                docTree = self.generate_tree(
                    target, req, node)
            elif rstFile:
                docTree = core.publish_doctree(
                    node.get_content().read())
            else:
                docTree = self.generate_docs(
                    target, req, node)
        except:
            docTree = self.generate_error(target, req)
                        
        content = core.publish_from_doctree(
            docTree, writer_name = "html")
        content = content.split("<body>", 1)[1]
        content = content.rsplit("</body>",1)[0]
        req.hdf['rstdoc.content'] = Markup(content)
        return content
    
    def generate_error(self, target, req):
        req.hdf['title'] = "RstDoc Error"
        return astdoc.documentError("Error processing '%s'" % 
            target)
        
    def generate_tree(self, target, req, node):
        title = "Subversion Directory for %s " % target

        def formatDocString(paragraphs, path, name):
            """ The paragraphs argument is a list of 0 
            or more docutils.nodes.paragraph nodes
            as returned by the astdoc.docString function.
            This method  processes the first paragraph 
            to add the Module name and hyperlink.
            
            """
            
            # Create a hyperlink node.
            ref = nodes.reference("", name,
                refuri = self.env.href.rstdoc(
                    "/".join(path)))
            
            # No first paragraph so create one with
            # the module name and hyperlink.    
            if not paragraphs:
                paragraphs = [nodes.paragraph("",
                    "test Module: ", ref)]
                    
            # Otherwise insert name and hyperlink
            # into the first paragraph.
            else:
                paragraphs[0].insert(0, nodes.Text(": "))
                paragraphs[0].insert(0, ref)
                paragraphs[0].insert(0, nodes.Text("Module "))
                
            return paragraphs
         
        def tree(path, node):
            """ Recursively build tree. """

            def getEntries(node):
                """ For each entry in **node**, determine the 
                node type (module,  normal file, or directory) and
                yield a tuple containing the entry's type, name, 
                and doc string (for modules) as well as the entry
                itself.  A list of these tuples is sorted to 
                order the definition list items. """

                for entry in node.get_entries():
                    if entry.isfile:
                        # Parse file and return module doc string
                        # or None for ordinary files.
                        doc = astdoc.moduleDocString(
                            entry.get_content().read().rstrip(), 
                            self.docStringLength
                            )
                            
                        if doc is not None:
                            # Entry is a module.
                            typ = 0

                        else:
                            # Entry is an ordinary file.
                            typ = 1
                            
                    else:
                        # Entry is a directory.
                        typ = 2
                        doc = None
                            
                    yield typ, entry.name, doc, entry
                
            if node.isfile:
                raise ValueError("Method 'tree' must be called "
                                        "with a directory node")

            # Process sub items first so we can determine
            # if the branch is a package or just a 
            # normal sub-directory.
            branchType = "Directory "
            item = nodes.definition_list_item("")
            entries = list(getEntries(node))
            entries.sort()
            
            # Add each entry to definition list item.
            for typ, name, doc, entry in entries:
                # Modules.
                if typ == 0:
                    if name == "__init__.py":
                        branchType = "Package "
                
                    defin = formatDocString(
                            doc, path + [name], name)
                
                # Ordinary files.
                elif typ == 1:
                    ref = nodes.reference("", name,
                        refuri = self.env.href.browser(
                            "/".join(path + [name])
                        ))
                    defin = [nodes.paragraph("", "File: ", ref)]
                
                # Sub-directories (process recursively)
                else:
                    defin = [tree(path + [name], entry)]

                item += nodes.definition("", *defin)

            # Now we can insert the 'term' element
            # and return the branch.
            name = branchType + (node.name or "ROOT")
            item.insert(0, nodes.term("", name))
            return nodes.definition_list("", item)
            
        doc = new_document("")
        doc += tree([target], node)
        return doc

    def generate_docs(self, target, req, node):
        req.hdf['title'] = "Documentation for %s" \
            % target
            
        module = astdoc.documentString(
            node.get_content().read(), target)
        return module.docTree()

