# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Creates `RestructuredText`-based documentation for Python 
modules and docstrings by parsing `AbstractSyntaxTrees` generated by
the Python compiler.

"""

import sys, os, os.path, itertools
from glob import glob
from types import StringTypes, TupleType
from compiler import walk, parse
from textwrap import dedent
from traceback import format_exc

from docutils import core, nodes, utils
from docutils.core import publish_doctree as _publish_doctree
from docutils.core import publish_from_doctree as _publish_from_doctree
from docutils.nodes import Text
from docutils.utils import new_document

from rstdoc.misclibs.util import *
from rstdoc.misclibs.functional import *
from rstdoc.rstlib.astwriter import nodeToString, getSignature

def overrideSettings(kw):
    settings_overrides = kw.get("settings_overrides", {})
    if "report_level" not in settings_overrides:
        settings_overrides["report_level"] = 9
        kw["settings_overrides"] = settings_overrides
    return kw
    
def publish_doctree(*args, **kw):
    kw = overrideSettings(kw)
    return _publish_doctree(*args, **kw)

def publish_from_doctree(*args, **kw):
    kw = overrideSettings(kw)
    return _publish_from_doctree(*args, **kw)

def documentError(title, includeTraceback = True):
    """ Format an exception.  
    
    >>> try:
    ...     raise ValueError("oops")
    ... except:
    ...     print publish_from_doctree(
    ...         documentError("Sample Error"))
    <document source="Sample Error">
        <error>
            <definition_list>
                <definition_list_item>
                    <term>
                        Sample Error
                    <definition>
                        <paragraph>
                            oops
                <literal_block xml:space="preserve">
                    Traceback (most recent call last):
                        ...
                    ValueError: oops
    <BLANKLINE>


    """
    
    node = nodes.definition_list("",
        nodes.definition_list_item("",
            nodes.term("", title),
            nodes.definition("", 
                nodes.paragraph("", str(sys.exc_info()[1])))))
    if includeTraceback:
        node += nodes.literal_block("", format_exc())
    node = nodes.error("", node)
    doc = new_document(title)
    doc += node
    return doc

def documentString(source, name = "<string>"):
    """ Create a tree of `NodeVisitor` subclasses.
    The instances have some usefull properties and
    the docTree method can be used to generate
    a docutils `document` tree which can be used to 
    generate HTML/LATEX documentation.
    
    source: string
        The source code of the module.
    name: string
        The name of the module to document (only
        used for reference purposes).
    returns: Module
        The returned object is an instance of Module.
        
    """
    
    return Module(source, name)

def documentFile(name):
    """ I perform the same operations as `documentModule`
    except I am passed a filename instead of the module
    source.
    
    name: string
        The filename of the module.  This file is 
        read to obtain the source for the module.
    returns: Module
        The returned object is an instance of Module.
        
    """
    return documentString(open(name).read(), name)
   
test1 = "- This is Junk -"
test2 = """
''' module.py

This is documentation.

This is too.
'''
"""

def moduleDocString(source, n=0, name = None):
    """ I attempt to parse the module docString from 
    source.  If any error occurs then None is returned.
    Otherwise I will return the first n paragraphs of the
    modules docstring. If n is 0 then all paragraphs are
    returned.
    
    >>> print moduleDocString(test1)
    None
    
    >>> print moduleDocString(test2)
    [<paragraph: <#text: u'module.py'>>, 
    <paragraph: <#text: u'This is document ...'>>, 
    <paragraph: <#text: u'This is too.'>>]

    >>> print moduleDocString(test2, 2)
    [<paragraph: <#text: u'module.py'>>, 
    <paragraph: <#text: u'This is document ...'>>]
    
    """
    
    class ModuleVisitor(object):
        docString = property(
            fget = lambda self:
                docString(self._docstring, n)
                )
        
        def __init__(self, source):
            self._docstring = ""
            node = parse(source)
            walk(node, self)
            
        def visitModule(self, node):
            self._docstring = node.doc
            
    try:
        return ModuleVisitor(source).docString
        
    except:
        return None

_filter = None
def setDocStringFilter(filter):
    """ Use me to define a function that it is applied
    to docstrings before they are processed.  
    
    >>> def caps(string): return string.upper()
    >>> setDocStringFilter(caps)
    >>> print docString("testme1\\n\\n\\ntestme2")
    [<paragraph: <#text: u'TESTME1'>>, <paragraph: <#text: u'TESTME2'>>]
    
    >>> setDocStringFilter(None)
    >>> print docString("testme1\\n\\n\\ntestme2")
    [<paragraph: <#text: u'testme1'>>, <paragraph: <#text: u'testme2'>>]
    
    """
    global _filter
    _filter = filter
 
def documentPackages(docPath, packages, writer_name = "html", 
        output = sys.stderr, title = "Package Index", 
        settings = {}):
            
    """ Document one or more *packages*, writing documentaion
    to *docPath* with docutils writer *writer_name*. Status
    messages are printed to file object *output*, set *output*
    to None to suppress messages. The *title* argument sets
    the doctree["title"] attribute and also sets the 
    heading of the index.
    
    """

    def defList(term, *defs):
        """ Create a definition list """
        
        return nodes.definition_list("",
            nodes.definition_list_item("",
                nodes.term("", term), 
                *[nodes.definition("", def_) for def_ in defs]
                ))
    
    def addModule(moduleDocs, docs):
        moduleDocString = \
            docString(moduleDocs.docstring)
        moduleDocString[0].insert(0, Text(": "))
        moduleDocString[0].insert(0, 
                nodes.reference("", 
                    "%s" % moduleName, 
                    refuri="%s" % moduleDocPath
                    )
                )
        moduleDocString[0].insert(0, Text("Module "))
        docs.append(moduleDocString[0])
        
        title = ("Module %s.%s" 
            % (package, moduleName)
            )
        moduleDocTree = moduleDocs.docTree()
        moduleDocTree.insert(0, 
            nodes.title("", title)
            )
        moduleDocTree['title'] = title
        backlink = nodes.paragraph("", "Return to ")
        backlink += nodes.reference("", 
            "index", refuri="index.%s" % writer_name)
        moduleDocTree.insert(0, backlink)
        writeDocs(moduleDocPath, moduleDocTree)

    def addFile(moduleName, docs):
        fileDocs = nodes.paragraph("","File ")
        fileDocs += nodes.strong("", moduleName)
        docs.append(fileDocs)
    
    def writeDocs(path, docTree):
        open(path, "w").write(
            core.publish_from_doctree(
                docTree,
                writer_name = writer_name,
                settings_overrides = settings_overrides,
                )
            )

    # Initialize setting to defaults.
    settings_overrides = {
            "report_level" : 9,
            "stylesheet_path" : 
                os.path.join(
                    os.path.dirname(__file__),
                    "package_docs.css"
                    )
            }
    # Update settings from the **setting** argument.
    settings_overrides.update(settings)

    if output:
        print >>output, "Documention directory: %s" % docPath
        
    docTree = new_document(docPath)
    docTree += nodes.title("", title)
    docTree['title'] = title
    
    for package in packages:
        if package == ".":
            pattern = "*"
            package = os.path.basename(os.getcwd())
        else:
            pattern = os.path.join(
                package.replace(".", os.path.sep),"*")
                
        docs = []; isPackage = False
        for module in filter(os.path.isfile, glob(pattern)):
            moduleName, ext = os.path.splitext(
                os.path.basename(module))
                
            if moduleName == "__init__":
                isPackage = True
                
            if ext in [".pyc", ".pyo"]:
                continue
                
            moduleDocName = "%s-%s.%s" % (
                    package.replace(".", "-"), 
                    moduleName, 
                    writer_name
                    )
            moduleDocPath = os.path.join(docPath, moduleDocName)
            if output:
                print >>output, "Processing: %s" % module

            if ext == ".py":
                try:
                    moduleDocs = documentFile(module)
                    
                except:
                    addFile(module, docs)
                    
                else:
                    addModule(moduleDocs, docs)
            else:
                addFile(module, docs)
            
        defTerm = "%s %s: " % \
            ((isPackage and "Package" or "Directory"), package)
        docTree += defList(defTerm, *docs)
    
    if output:
        print >>output, "Generating index.%s" % writer_name
        
    indexPath = os.path.join(
        docPath.replace(".","-"), "index.%s" % writer_name)
    writeDocs(indexPath, docTree)

 
def docString(docstring, n=0, signature = None):
    """ I return a list of docutils paragraph nodes 
    for the specified docstring.  The '''`n`''' 
    argument specifies the maximum number of paragraphs
    that will be returned.  

    Use the method setDocString filter to "wikify" or 
    otherwise process the docstring by replacing string 
    patterns with reStructured text roles or other text.
    
    >>> print docString("")
    [<paragraph: >]
    
    >>> print docString("testme1\\n\\n\\ntestme2")
    [<paragraph: <#text: u'testme1'>>, 
    <paragraph: <#text: u'testme2'>>]

    >>> print docString("testme1\\n\\n\\ntestme2", 1)
    [<paragraph: <#text: u'testme1'>>]

    
    """
    
    docstring = docstring or ""
    if _filter is not None:
        docstring = _filter(docstring)
        
    lines = docstring.splitlines()
    first = lines and lines.pop(0).strip() or ""
    rest = dedent("\n".join(lines))
    if signature:
        signature = signature.replace('*', '\\*')
        docstring = "(%s)\n" % signature + first + "\n" + rest
        n += 1
    else:
        docstring = first + "\n" + rest

    paragraphs = publish_doctree(docstring).children
    
    if not paragraphs:
        paragraphs = [nodes.paragraph("")]
    n = n or len(paragraphs)
    return paragraphs[:n]
        
class NodeVisitor(object):
    """ I am the base for all visitor subclasses. These
    classes should not be created individually.  Use
    the `documentFile` or `documentString` methods instead.
            
    The only public method of this class is 
    `docTree`.   The remaining methods are called
    by the node walker.
    """
    
    _typeName = None

    class typeName(Property):
        """ The type of node: Module, Class,... """
        
        fset = fdel = None
        
        def fget(self, instance):
            return instance._typeName

    class name(Property):
        """ The name of the node. """
        
    class lineno(Property):
        """ The source line defining the node. """

    class signature(Property):
        """ The bases/arguments used to define the node. """
        
    class docstring(Property):
        """ The node's docstring. """

    class nodeMap(Property):
        """ Maps nodes to names."""
        
    hasChildren = property(
        fget = lambda self: self._hasChildren())
    hasDocstring = property(
        fget = lambda self: self._hasDocstring())
    imports = property(
        fget = lambda self: self._get("_imports"))
    classes = property(
        fget = lambda self: self._get("_classes"))
    methods = property(
        fget = lambda self: self._get("_methods"))
    attributes = property(
        fget = lambda self: self._get("_attributes"))
    globals = property(
        fget = lambda self: self._get("_globals"))
    decorators = property(
        fget = lambda self: self._get("_decorators"))
    
    def __init__(self, name, lineno = None, 
        signature = None, docstring = None, node = None):
        
        """ Create a NodeVisitor instance. 
        
        name: string
            The name of the module to document (only
            used for reference).
        lineno: int
            The line number of the source which defines
            the node.
        signature: string
            The `signature` of the node, def foo(bar),
            for example.
        docstring: string
            The nodes docstring.
        node: node
            The Ast node of the current item. See the
            standard library compiler module.
        """
        
        self.name = name
        self.lineno = lineno
        self.docstring = docstring or ""
            
        self._imports = []
        self._classes = []
        self._methods = []
        self._attributes = []
        self._globals = []
        self._decorators = []
        self._nodeMap = {}
        
        if signature is None:   
            self.signature = \
                "%s %s" % (self.typeName, self.name)
        else:
            self.signature = signature
            
        if node:
            walk(node, self)
            
    def _hasChildren(self):
        """ Returns True if the node has any children.
        """
        
        return bool(self._classes
            or self._methods 
            or self._imports
            or self._attributes)

    def _hasDocstring(self):
        """ Returns True if the node has any children.
        """
        
        return len(docString(self.docstring)) > 1

    def visitModule(self, node):
        self.docstring = node.doc
        walk(node.node, self)
    
    def _get(self, attr):
        def include(item):
            return bool(item.docstring is None 
                    or item.docstring.strip()
                    or item.hasChildren)
                                        
        #items = filter(include, getattr(self, attr)[:])
        items = getattr(self, attr)[:]
        items.sort(key = attrgetter("name"))
        return items
        
    def visitClass(self, node):
        self._classes.append(Class(node))
        
    def visitFunction(self, node):
        self._methods.append(Method(node))

    def visitImport(self, node):
        for name in node.names:
            if name[0] not in [n.name 
                    for n in self._imports]:
                self._imports.append(
                    Import(name[0], node.lineno))
        
    def visitFrom(self, node):
        if node.modname not in [n.name
                for n in self._imports]:
            self._imports.append(
                Import(node.modname, node.lineno))
            
    def visitGlobal(self, node):
        for name in node.names:
            self._globals.append(
                Global(name, node.lineno))
                
    def visitAssign(self, node):
        targets = [nodeToString(n) for n in node.nodes]
        name = "%s = ..." % " = ".join(targets)
        docstring = "%s = %s" % (" = ".join(targets),
            nodeToString(node.expr))

        self._attributes.append(
            Attribute(name, node.lineno, docstring))
         
    def docTree(self):
        """ Create a docutils document tree for the
        current instance. 
        """
        
        doc = new_document(self.name)
        doc += self._docTree()
        return doc
        
    def _docTree(self, prefix = ""):
        """ I create a section for the current node,
        recursively calling all children as required.
        
        prefix: string
            The part of the full dotted name prior to 
            the current node.  For example, if the current 
            node is myMethod of MyClass in themodule then
            the prefix would be "themodule.MyClass.".
            
        """
        
        self._nodeMap = {}
        prefix = prefix or ""
        nextID = itertools.count(1).next
        
        def paragraph(text):
            """ I create a definition list definition
            with a paragraph of text. """
            
            return nodes.definition("", 
                nodes.paragraph("",text))
                    
            
        def link(node):
            """ I create a definition list definition 
            with a link. """
            
                    
            return paragraphs[0]
                
        def defList(term, *defs):
            """ I create a definition list. """
            
            return nodes.definition_list("",
                nodes.definition_list_item("",
                    nodes.term("", term), *defs))

        def definition(node):
            paragraphs = docString(node.docstring, 1)
            if node.typeName == "Class" and not node.signature:
                signature = ""
            else:
                signature = "(%s)" % node.signature
            
            paragraphs[0].insert(0, nodes.Text(": "))
            if node.hasChildren or node.hasDocstring:
                id = prefix+self.name + "." + node.name
                self._nodeMap[id] = node
                paragraphs[0].insert(0, 
                    nodes.reference("", 
                        "%s%s" % (node.name, signature),
                        refid="%s" % id
                    ))
                
            else:
                paragraphs[0].insert(0, 
                    nodes.strong("", 
                        "%s%s" % (node.name, signature) 
                    ))
                    
                
            return nodes.definition("", *paragraphs)
                    
        # Create document tree with back link.
        doc = nodes.section(
            ids = ["%s%s" % (prefix, self.name)]
            )
        doc += nodes.transition("")
        doc += nodes.title("", 
            "%s %s%s" % (self.typeName, prefix, self.name),
            refname = "%s%s" % (prefix, self.name)
            )
            
        if prefix:
            ref = prefix[:-1]
            doc += nodes.paragraph("","",
                nodes.reference("","Back to %s" % ref, 
                refid = ref
                ))
                
        # Add the item signature.
        if self.typeName == "Class":
            signature = self.signature or "None"
            doc += defList("Bases:", paragraph(signature))
        elif self.typeName == "Method":
            doc += defList("Arguments:", paragraph(self.signature))
        
        if self.docstring:
            doc += docString(self.docstring)

        # Process Decorators:
        if self.decorators:
            doc += defList("Decorators:", 
                paragraph(", ".join([node.signature 
                    for node in self.decorators])))
            
        # Process imports.
        if self.imports:
            doc += defList("Imports:", 
                paragraph(", ".join([node.name 
                    for node in self.imports])))
        
        subDocs = []
        # Process classes.
        if self.classes:
            doc += defList("Classes:", 
                *[definition(node) for node in self.classes])
            
        # Process methods.
        if self.methods:
            doc += defList("Methods:", 
                *[definition(node) for node in self.methods])
            
        # Process attributes.
        if self.attributes:
            doc += defList("Attributes:", 
                *[paragraph(node.docstring) 
                    for node in self.attributes])
        
        # Process globals:
        if self.globals:
            doc += defList("Globals:", 
                paragraph(", ".join([node.name 
                    for node in self.globals])))

        subDocs = itertools.chain(
            self.classes, self.methods)
        for item in subDocs:
            if item.hasChildren or item.hasDocstring:
                doc += item._docTree(prefix+self.name+".")
        return doc

class Module(NodeVisitor):
    """ An instance of Module is created at the top level
    of the node tree.   Use the `documentString` or 
    `documentFile` methods to create me.
    """
    
    _typeName = "Module"
    
    def __init__(self, source, name = "<string>"):
        """ Create a Module instance.
            
            source: string
                The source code for the module.
            name: string
                The name of the module for reference only.
        """
        source = source.rstrip()
        fileName = os.path.splitext(
            os.path.basename(name))[0]
        super(Module, self).__init__(
            name = fileName, 
            lineno = None, 
            signature = fileName, 
            node = parse(source)
            )
    
        
class Import(NodeVisitor):
    """ An Import instance is created for each imported
    module. 
    """
    
    _typeName = "Import"
    
    def __init__(self, name, lineNo):
        """ Create an Import instance.
            
            name: string
                The name of the imported module.
            lineNo: int
                The line number of the import/from statement.
        """
        
        super(Import, self).__init__(name, lineNo, name)


class Global(NodeVisitor):
    """ A Global instance is created for each global
    statement.
    """
    
    _typeName = "Global"
    
    def __init__(self, name, lineNo):
        """ Create an Global instance.
            
            name: string
                The name of the global variable.
            lineNo: int
                The line number of the global statement.
        """
        
        super(Global, self).__init__(name, lineNo, name)

        
class Class(NodeVisitor):
    """ A Class instance is created for each class 
    definition.
    """
    
    _typeName = "Class"
    
    class bases(Property):
        """ The class bases. """
    
    def __init__(self, node):
        """ Create a Class instance.
        
        node: node
            The node of the class definition.
        """
        
        self.bases = bases = \
            [nodeToString(base) for base in node.bases]
            
        if bases:
            signature = ", ".join(bases)
            #name = "%s(%s)" %(node.name, signature)
            
        else:
            signature = ""
        name = node.name
        super(Class, self).__init__(name, node.lineno, 
            signature, node.doc, node.code)

class Method(NodeVisitor):
    """ A Method instance is created for each method 
    definition. 
    """
    
    _typeName = "Method"
    
    def __init__(self, node):
        """ I create a Method instance.
        
        node: compiler.ast.node
            The node instance.
        
        """
        args = ", ".join(getSignature(node))
        signature = "%s" % (args)
        name = "%s(%s)" % (node.name, args)
        
        super(Method, self).__init__(
            node.name, node.lineno, 
            signature, node.doc, node.code)
        if node.decorators:
            for deco in node.decorators:
                self._decorators.append(Decorator(deco))

    def visitAssign(self, node):
        """ I am a no-op (we do not want to see
        assignments to local variables.)
        """
        pass

class Decorator(NodeVisitor):
    """ A Decorator instance is created for each decorater
    of a method.
    """
    
    
    _typeName = "Decorator"
    
    def __init__(self, node):
        """ I create a Method instance.
        
        node: node
            The node of the def statement.
        """
        
        signature = "@" + nodeToString(node)
        super(Decorator, self).__init__(
            signature, node.lineno, signature)

class Attribute(NodeVisitor):
    """ An Attribute instance is created for each 
    top-level assignment statement.
    """
    
    _typeName = "Attribute"
    
    def __init__(self, name, lineNo, docstring):
        """ I create an Attribute instance.
        
        name: string
            The attribute name.
        lineNo: int
            The line number where the assignment was made.
        docstring: string
            A docstring is created from the entire 
            assignment statement.  
            
        """
        super(Attribute, self).__init__(
            name, lineNo, name, docstring)

def testSuite():
    import doctest as dt
    import astdoc as moduleToTest
    
    return dt.DocTestSuite(moduleToTest, 
        optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
            dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)
    
if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())
