# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Generate Python source from `AbstractSyntaxTrees` generated
by the compiler. """

from types import StringTypes, TupleType
from cStringIO import StringIO
from compiler import walk, parse, ast

def nodeToString(ast):
    """ I convert an ast tree to source. """

    sw = SourceWriter()
    walk(ast, sw)
    return sw.s.getvalue()

def sourceToString(source):
    """ I convert a source string to a source string
    by creating an ast tree and then applying nodeToString.
    """

    return nodeToString(parse(source))

def fileToString(name):
    """ I convert a source file to a source string
    by creating an ast tree and then applying nodeToString.
    """

    return sourceToString(open(name).read())

def getSignature(node):
    """ I return the argument signature for node
    which must be a Function or Lambda node.
    """

    assert isinstance(node, ast.Function) or \
            isinstance(node, ast.Lambda), \
            "Invalid node type: %s" % node

    def fix(arg):
        if isinstance(arg, StringTypes):
            return arg
        else:
            assert isinstance(arg, TupleType)
            args = arg
            return "(%s)" % ",".join(
                [fix(arg) for arg in args])

    argnames = map(fix, node.argnames)
    defaults = map(nodeToString, node.defaults)
    starStar = bool(node.flags & 0x8)
    star = bool(node.flags & 0x4)

    # First handle (..., *args) and/or (..., **kw)
    if starStar and star:
        varArgs = [
            "*%s" % argnames[-2],
            "**%s" % argnames[-1]]
    elif starStar:
        varArgs = ["**%s" % argnames[-1]]
    elif star:
        varArgs = ["*%s" % argnames[-1]]
    else:
        varArgs = []
    if varArgs:
        argnames = argnames[:-len(varArgs)]

    # Next handle arguments that have defaults.
    if defaults:
        argsWithDefaults = argnames[-len(defaults):]
        argnames = argnames[:-len(defaults)]
    else:
        argsWithDefaults = []

    # Whats left have no default values.
    argsWithoutDefaults = argnames

    # Put them all together and we are done.
    return argsWithoutDefaults + \
        map(lambda arg,default: "%s=%s" % (arg,default),
            argsWithDefaults, defaults) + \
        varArgs

class SourceWriter(object):
    """ Walk with me to generate source:

        >>> source = "1+2"
        >>> ast = parse(source)
        >>> print walk(ast, SourceWriter())
        1 + 2
        <BLANKLINE>


    Normally this class is not used directly, use
    the fileToString, nodeToString, or sourceToString
    methods instead.

    """

    _i = 0

    def __init__(self):
        """ Creates a SourceWriter instance. """

        self.s = StringIO()

    def w(self, s):
        """ Write to internal buffer. """

        self.s.write(s)

    def nl(self):
        """ Write a newline to the internal buffer and
        then indent to the current indentation level. """

        self.s.write('\n')
        self.s.write(' ' * 4 * self._i)

    def indent(self):
        """ Increase the indentation level. """

        self._i += 1
        self.nl()

    def dedent(self):
        """ Decrease the indentation level. """

        self._i -= 1
        self.nl()

    def joinNodes(self, join, nodeSeq, nodeWriter = None):
        """ I act like string.join(seq) except seq is
        a list of nodes, and the results are written to
        the internal buffer and not returned by this
        function.

        """

        if nodeWriter is None:
            def nodeWriter(node):
                walk(node, self)
        if nodeSeq:
            for node in nodeSeq[:-1]:
                nodeWriter(node)
                self.w(join)
            nodeWriter(nodeSeq[-1])

    def writeDoc(self, docString):
        """ I write the source for a docstring to the
        internal buffer.

        """

        if docString:
            self.w('''"""%s''' % docString)
            self.nl()
            self.w('''"""''')
            self.nl()

    def visitAdd(self, node):
        """ Binary addition, i.e. "+".

        >>> print sourceToString("1+2")
        1 + 2
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' + ')
        walk(node.right, self)

    def visitAnd(self, node):
        """ Called for logical and operations.  Note that
        a series of and's will result in only one to this
        method.

        >>> print sourceToString("1 and 2 and 3")
        1 and 2 and 3
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        self.joinNodes(" and ", node.nodes)

    def visitAssAttr(self, node):
        """ Assign expression to an attribute.

        >>> print sourceToString("test.a=5")
        test.a = 5
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.expr, self)
        self.w('.')
        self.w(node.attrname)

    def visitAssList(self, node):
        """ Assign expression to a list.

        >>> print sourceToString("[a,b,c] = [1,2,3]")
        [a, b, c] = [1, 2, 3]
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('[')
        self.joinNodes(", ", node.nodes)
        self.w(']')

    def visitAssName(self, node):
        """ Assign expression to a variable.

        >>> print sourceToString("x,y,z = 1, str(2), False")
        (x, y, z) = (1, str(2), False)
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        if node.flags == 'OP_DELETE':
            self.w("del ")
        self.w(node.name)

    def visitAssTuple(self, node):
        """ Assign expression to a tuple.

        >>> print sourceToString("x, (a,b) ,z = 1, (str(2),str(3)), False")
        (x, (a, b), z) = (1, (str(2), str(3)), False)
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('(')
        self.joinNodes(", ", node.nodes)
        self.w(')')

    def visitAssert(self, node):
        """ The assert statement.

        >>> print sourceToString("assert True == False, 'Oops'")
        assert (True == False, 'Oops')
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w("assert (%s" % nodeToString(node.test))
        if node.fail is None:
            self.w(")")
        else:
            self.w(", %s)" % nodeToString(node.fail))

    def visitAssign(self, node):
        """ Called for each assignment.  Note that
        a statement like a = b = None will result in
        one call to this method.

        >>> print sourceToString("x,y,z = 1, str(2), False")
        (x, y, z) = (1, str(2), False)
        <BLANKLINE>

        >>> print sourceToString("x,y,z = l = 1, str(2), False")
        (x, y, z) = l = (1, str(2), False)
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.joinNodes(" = ", node.nodes)
        self.w(' = ')
        walk(node.expr, self)
        self.nl()

    def visitAugAssign(self, node):
        """ Called for operations like x += 5 and x -= 1.

        >>> print sourceToString("x += 5")
        x += 5
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.node, self)
        self.w(' ')
        self.w(node.op)
        self.w(' ')
        walk(node.expr, self)
        self.nl()

    def visitBackquote(self, node):
        """ Called when the backquote character is used.

        >>> print sourceToString("`a`")
        `a`
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('`')
        walk(node.expr, self)
        self.w('`')

    def visitBitand(self, node):
        """ Called for bitwise 'and' operations. Note that
        a series of "&"'s will  result in only one call
        to this method.

        >>> print sourceToString("1&2&3")
        1 & 2 & 3
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w(" & ".join(
            [nodeToString(node) for node in node.nodes]))

    def visitBitor(self, node):
        """ Called for bitwise 'or' operations. Note that
        a series of "|"'s will  result in only one call
        to this method.

        >>> print sourceToString("1|2|3")
        1 | 2 | 3
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.joinNodes(" | ", node.nodes)

    def visitBitxor(self, node):
        """ Called for bitwise 'xor' operations. Note that
        a series of "^"'s will  result in only one call
        to this method.

        >>> print sourceToString("1^2^3")
        1 ^ 2 ^ 3
        """

        #print "Junk" # Uncomment to test the doc test
        self.joinNodes(" ^ ", node.nodes)

    def visitBreak(self, node):
        """ Called for a break statement.

        >>> print sourceToString("if True: break")
        if True:
            break
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w("break")

    def visitCallFunc(self, node):
        """ Called for a function call.

        >>> print sourceToString("x,y,z = 1, str(2), False")
        (x, y, z) = (1, str(2), False)
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.node, self)
        self.w('(')
        self.joinNodes(", ", node.args)
        self.w(')')


    def visitClass(self, node):
        """ Called for a class definition.

        >>> source = '''
        ... class Test(Base1, Base2):
        ...     \"\"\" Documentation string
        ...     This is a test. \"\"\"
        ...     pass
        ... '''
        >>>
        >>> print sourceToString(source)
        class Test(Base1, Base2):
            \"\"\" Documentation string
            This is a test.
            \"\"\"
        <BLANKLINE>
            pass
        <BLANKLINE>

        >>> source = '''
        ... class Test:
        ...     \"\"\" Documentation string
        ...     This is a test. \"\"\"
        ...     pass
        ... '''
        >>>
        >>> print sourceToString(source)
        class Test:
            \"\"\" Documentation string
            This is a test.
            \"\"\"
        <BLANKLINE>
            pass
        <BLANKLINE>


        """

        #print "Junk" # Uncomment to test the doc test
        self.w('class ')
        self.w(node.name)
        if node.bases:
            self.w('(')
            self.joinNodes(", ", node.bases)
            self.w(')')
        self.w(':')
        self.indent()

        try:
            if node.doc is not None:
                self.writeDoc(node.doc)

            self.nl()
            walk(node.code, self)

        finally:
            self.dedent()

    def visitCompare(self, node):
        """ Called for each comparison operation.

        >>> print sourceToString("a > b or c <= d")
        a > b or c <= d
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.expr, self)
        for (op, arg) in node.ops:
            self.w(' ')
            self.w(op)
            self.w(' ')
            walk(arg, self)

    def visitConst(self, node):
        """ Called for constants.

        >>> print sourceToString("1 + 2; 'A string constant'")
        1 + 2
        'A string constant'
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w(repr(node.value))

    def visitContinue(self, node):
        """ Called for the continue statement.

        >>> print sourceToString("if True: continue")
        if True:
            continue
        """

        #print "Junk" # Uncomment to test the doc test
        self.w("continue")

    def visitDecorators(self, node):
        """ Called for function decorators. Note if
        multiple decorators are applied to a function,
        this method is called only once.

        >>> source = '''
        ... @deco
        ... @deco(1,2,3)
        ... def test(): pass
        ... '''
        >>> print sourceToString(source)
        @deco
        @deco(1, 2, 3)
        def test():
            pass
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        for deco in node.nodes:
            self.w("@")
            walk(deco, self)
            self.nl()


    def visitDict(self, node):
        """ Called for a literal dictionary definition
        (i.e. { a : b, c : d })

        >>> print sourceToString("{a:5,'t':test}")
        { a : 5, 't' : test }

        """

        def nodeWriter((k,v)):
            walk(k, self)
            self.w(' : ')
            walk(v, self)

        #print "Junk" # Uncomment to test the doc test
        self.w('{ ')
        self.joinNodes(", ", node.items, nodeWriter)
        self.w(' }')

    def visitDiscard(self, node):
        """ This is called when the results of
        an expression are discarded.

        >>> print sourceToString("a")
        a
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.expr, self)
        self.nl()

    def visitDiv(self, node):
        """ Called for each division operation.

        >>> print sourceToString("1/2")
        1 / 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' / ')
        walk(node.right, self)

    def visitEllipsis(self, node):
        """ I don't know why I would want it but...

        >>> print sourceToString("a[...]")
        a[...]
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('...')

    def visitExec(self, node):
        """ Called for the exec statement.

        >>> print sourceToString("exec 'a=1'in {}, {}")
        exec 'a=1' in {  } , {  }
        >>> print sourceToString("exec 'a=1'in {}")
        exec 'a=1' in {  }
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('exec ')
        walk(node.expr, self)
        self.w(" in ")
        walk(node.locals, self)
        if node.globals is not None:
            self.w(" , ")
            walk(node.globals, self)

    def visitFloorDiv(self, node):
        """ Called for floor division.

        >>> print sourceToString("1//2")
        1 // 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' // ')
        walk(node.right, self)

    def visitFor(self, node):
        """ Called for a for statement.

        >>> print sourceToString('''
        ... for x in [1]:
        ...     print x
        ... else:
        ...     print "else"
        ... ''')
        for x in [1]:
            print x
        else:
            print 'else'
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('for ')
        walk(node.assign, self)
        self.w(' in ')
        walk(node.list, self)
        self.w(':')
        self.indent()
        try:
            walk(node.body, self)
        finally:
            self.dedent()
        if node.else_:
            self.w('else:')
            self.indent()
            try:
                walk(node.else_, self)
            finally:
                self.dedent()

    def visitFrom(self, node):
        """ Called for a from (import) statement.

        >>> print sourceToString('''
        ... from module import some as thing, a as b
        ... ''')
        from module import some as thing, a as b
        """

        def nodeWriter((mod, _as)):
            self.w(mod)
            if _as is not None:
                self.w(' as ')
                self.w(_as)

        #print "Junk" # Uncomment to test the doc test
        self.w('from ')
        self.w(node.modname)
        self.w(' import ')
        self.joinNodes(", ", node.names, nodeWriter)
        self.nl()


    def visitFunction(self, node):
        """ Called for each function definition.

        >>> print sourceToString('''
        ... def test(a,b=1, *a, **k):
        ...     \"\"\" This is my docstring.
        ...     Some more documentation.\"\"\"
        ...     pass
        ... ''')
        def test(a, b=1, *a, **k):
        <BLANKLINE>
            \"\"\" This is my docstring.
            Some more documentation.
            \"\"\"
            pass
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        if node.decorators:
            walk(node.decorators, self)
        self.w('def %s(%s):' % (
            node.name, ", ".join(getSignature(node))))
        self.indent()
        if node.doc:
            self.nl()
            self.writeDoc(node.doc)
        try:
            walk(node.code, self)
        finally:
            self.dedent()

        self._decorators = node.decorators

    def visitGetattr(self, node):
        """ Called for attribute access.

        >>> print sourceToString("a = test.a")
        a = test.a
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.expr, self)
        self.w('.')
        self.w(node.attrname)

    def visitGlobal(self, node):
        """ Called for the global statement.

        >>> print sourceToString("global a,b,c")
        global a, b, c

        """

        #print "Junk" # Uncomment to test the doc test
        self.w("global %s" % ", ".join(node.names))

    def visitIf(self, node):
        """ Called for the if statement.  Also handles
        the elif and else clauses.

        >>> source = '''
        ... if True:
        ...     print "true"
        ... elif False:
        ...     print "false"
        ... else:
        ...     print "I am confused"
        ... '''
        >>> print sourceToString(source)
        if True:
            print 'true'
        <BLANKLINE>
        elif False:
            print 'false'
        <BLANKLINE>
        else:
            print 'I am confused'
        <BLANKLINE>
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        keyword = 'if'
        for (cond, body) in node.tests:
            self.w(keyword)
            self.w(' ')
            walk(cond, self)
            self.w(':')
            self.indent()
            try:
                walk(body, self)
            finally:
                self.dedent()
            keyword = 'elif'
        if node.else_:
            self.w('else:')
            self.indent()
            try:
                walk(node.else_, self)
            finally:
                self.dedent()

    def visitImport(self, node):
        """ Called for the import statement.

        >>> print sourceToString(
        ...     "import a.b as b, some.thing as some")
        import a.b as b, some.thing as some
        """
        def nodeWriter((mod, _as)):
            self.w(mod)
            if _as is not None:
                self.w(' as ')
                self.w(_as)

        #print "Junk" # Uncomment to test the doc test
        self.w('import ')
        self.joinNodes(", ", node.names, nodeWriter)
        self.nl()

    def visitInvert(self, node):
        """ Called for bitwise inversion.

        >>> print sourceToString("~ 5")
        ~5
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('~')
        walk(node.expr, self)

    def visitKeyword(self, node):
        """ Called for a keyword argument.

        >>> print sourceToString("f(a=5)")
        f(a = 5)
        <BLANKLINE>
        """

        #print "Junk" # Uncomment to test the doc test
        self.w("%s = " % node.name)
        walk(node.expr, self)

    def visitLambda(self, node):
        """ Called for a lambda statement.

        >>> print sourceToString("lambda a,b=1, *a, **k: a")
        lambda a, b=1, *a, **k: a

        """

        #print "Junk" # Uncomment to test the doc test
        self.w('lambda %s: '
            % ", ".join(getSignature(node)))
        walk(node.code, self)

    def visitLeftShift(self, node):
        """ Called for bitwise shift-left.

        >>> print sourceToString("1<<2")
        1 << 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' << ')
        walk(node.right, self)

    def visitList(self, node):
        """ Called for a literal list.

        >>> print sourceToString("[a,b,c]")
        [a, b, c]
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('[')
        self.joinNodes(", ", node.nodes)
        self.w(']')

    def visitListComp(self, node):
        """ Called for a list comprehension.

        >>> source = '[x for x in range(10)]'
        >>> print sourceToString(source)
        [x for x in range(10)]

        """

        #print "Junk" # Uncomment to test the doc test
        self.w('[')
        walk(node.expr, self)
        for q in node.quals:
            walk(q, self)
        self.w(']')

    def visitListCompFor(self, node):
        """ Hmmm, if expr1 if expr2 is valid syntax
        (at least in 2.4.2.)

        >>> source = '[x for x in range(10) if x if y]'
        >>> print sourceToString(source)
        [x for x in range(10) if x if y]

        """

        #print "Junk" # Uncomment to test the doc test
        self.w(' for ')
        walk(node.assign, self)
        self.w(' in ')
        walk(node.list, self)
        for expr in node.ifs:
            self.w(' if ')
            walk(expr, self)

    def visitListCompIf(self, node):
        """ Used for the if part of a list comprehension.

        >>> source = '[x for x in range(10) if x if y]'
        >>> print sourceToString(source)
        [x for x in range(10) if x if y]

        """

        #print "Junk" # Uncomment to test the doc test
        #self.w(' if ')
        walk(node.test, self)

    def visitMod(self, node):
        """ Called for modulus division.

        >>> print sourceToString("1%2")
        1 % 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' % ')
        walk(node.right, self)

    def visitModule(self, node):
        """ This allways gets visited first.

        >>> print sourceToString("' test'")
        \"\"\" test
        \"\"\"
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        if node.doc is not None:
            self.writeDoc(node.doc)
        walk(node.node, self)

    def visitMul(self, node):
        """ Called for the multiplication operation.

        >>> print sourceToString("1*2")
        1 * 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' * ')
        walk(node.right, self)

    def visitName(self, node):
        """ Called for variable access.

        >>> print sourceToString("test")
        test
        """

        #print "Junk" # Uncomment to test the doc test
        self.w(node.name)

    def visitNot(self, node):
        """ Called for logical not.

        >>> print sourceToString("not True")
        not True
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('not ')
        walk(node.expr, self)

    def visitOr(self, node):
        """ Called for logical or.  Note that multiple
        or's result in one call to this method.

        >>> print sourceToString("a or b or c")
        a or b or c
        """

        #print "Junk" # Uncomment to test the doc test
        self.joinNodes(" or ", node.nodes)

    def visitPass(self, node):
        """ Called for the pass statement.

        >>> print sourceToString("def test(): pass")
        def test():
            pass
        """

        #print "Junk" # Uncomment to test the doc test
        self.w("pass")

    def visitPower(self, node):
        """ Called for exponation oerations.

        >>> print sourceToString("1**2")
        1 ** 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' ** ')
        walk(node.right, self)

    def visitPrint(self, node):
        """ Called for a print statement without a
        trailing linefeed.

        >>> print sourceToString("print >>f,a,b,c,")
        print >>f, a, b, c,
        >>> print sourceToString("print a,b,c,")
        print a, b, c,

        """

        #print "Junk" # Uncomment to test the doc test
        self.w('print ')
        if node.dest:
            self.w('>>')
            walk(node.dest, self)
            self.w(', ')

        for e in node.nodes:
            walk(e, self)
            self.w(', ')

        self.nl()

    def visitPrintnl(self, node):
        """ Called for a print statement with a
        trailing linefeed.

        >>> print sourceToString("print >>f,a,b,c")
        print >>f, a, b, c
        >>> print sourceToString("print a,b,c")
        print a, b, c
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('print ')
        if node.dest:
            self.w('>>')
            walk(node.dest, self)
            self.w(', ')

        if node.nodes:
            for e in node.nodes[:-1]:
                walk(e, self)
                self.w(', ')

            walk(node.nodes[-1], self)

        self.nl()

    def visitRaise(self, node):
        """ Called for the raise statement.

        >>> print sourceToString(
        ...     "raise ValueError, value, tb")
        raise ValueError, value, tb
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('raise ')
        walk(node.expr1, self)
        if node.expr2:
            self.w(', ')
            walk(node.expr2, self)
            if node.expr3:
                self.w(', ')
                walk(node.expr3, self)
        self.nl()

    def visitReturn(self, node):
        """ Called for the return statement.

        >>> print sourceToString("return some,things")
        return (some, things)
        """

        #print "Junk" # Uncomment to test the doc test
        self.w('return ')
        walk(node.value, self)
        self.nl()

    def visitRightShift(self, node):
        """ Called for bitwise shift right.

        >>> print sourceToString("1>>2")
        1 >> 2
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' >> ')
        walk(node.right, self)


    def visitSliceobj(self, node):
        """ Called for sequence slice operations.

        >>> print sourceToString("l[a: b: c ]")
        l[a:b:c]
        <BLANKLINE>
        >>> print sourceToString("l[:b:c ]")
        l[:b:c]

        >>> print sourceToString("l[a::c ]")
        l[a::c]

        >>> print sourceToString("l[a:b:]")
        l[a:b]
        """

        #print "Junk" # Uncomment to test the doc test
        nodes = map(
            lambda node: node != "None" and node or "",
            [nodeToString(node) for node in node.nodes])
        while nodes[-1] == "":
            nodes = nodes[:-1]
        self.w(":".join(nodes))

    def visitStmt(self, node):
        """ Called for each statement.

        >>> print sourceToString("print x")
        print x

        """

        #print "Junk" # Uncomment to test the doc test
        for n in node.getChildren():
            walk(n, self)

    def visitSub(self, node):
        """ Called for each subtraction operation.

        >>> print sourceToString("x-1")
        x - 1
        """

        #print "Junk" # Uncomment to test the doc test
        walk(node.left, self)
        self.w(' - ')
        walk(node.right, self)

    def visitSubscript(self, node):
        """ Called for a subscript (not a slice).

        >>> print sourceToString("[1,2,3][0]")
        [1, 2, 3][0]
        >>> print sourceToString("del l[2]")
        del l[2]
        """

        #print "Junk" # Uncomment to test the doc test
        if node.flags == 'OP_DELETE':
            self.w("del ")

        walk(node.expr, self)
        self.w('[')
        walk(node.subs[0], self)
        self.w(']')

    def visitTryExcept(self, node):
        """ Called for a try/except block.

        >>> source = '''
        ... try:
        ...     some(thing)
        ... except ValueError,e:
        ...     done()
        ... '''
        >>> print sourceToString(source)
        try:
            some(thing)
        <BLANKLINE>
        except ValueError, e:
            done()
        <BLANKLINE>
        <BLANKLINE>

        >>> source = '''
        ... try:
        ...     some(thing)
        ... except (ValueError,CustomError),e:
        ...     done()
        ... '''
        >>> print sourceToString(source)
        try:
            some(thing)
        <BLANKLINE>
        except (ValueError, CustomError), e:
            done()
        <BLANKLINE>
        <BLANKLINE>
        """

        self.w('try:')
        self.indent()
        walk(node.body, self)
        self.dedent()
        for excType, excValue, handler in node.handlers:
            self.w('except')
            if excType:
                self.w(" ")
                walk(excType, self)
            if excValue:
                self.w(", ")
                walk(excValue, self)
            self.w(":")
            self.indent()
            walk(handler, self)
            self.dedent()


    def visitTryFinally(self, node):
        """ Called for a try/finally block.

        >>> source = '''
        ... try:
        ...     some(thing)
        ... finally:
        ...     done()
        ... '''
        >>> print sourceToString(source)
        try:
            some(thing)
        <BLANKLINE>
        finally:
            done()
        <BLANKLINE>
        <BLANKLINE>

        """

        #print "Junk" # Uncomment to test the doc test
        self.w('try:')
        self.indent()
        try:
            walk(node.body, self)
        finally:
            self.dedent()
        self.w('finally:')
        self.indent()
        try:
            walk(node.final, self)
        finally:
            self.dedent()

    def visitTuple(self, node):
        """ Called for each tuple.

        >>> print sourceToString("(1,2,3)")
        (1, 2, 3)
        >>> print sourceToString("(1,)")
        (1,)

        """

        self.w('(')
        if len(node.nodes) == 0:
            pass
        elif len(node.nodes) == 1:
            walk(node.nodes[0], self)
            self.w(',')
        else:
            self.joinNodes(", ", node.nodes)
        self.w(')')

    def visitUnaryAdd(self, node):
        """ Called for unary "+".

        >>> print sourceToString("+100")
        +100
        """

        self.w('+')
        walk(node.expr, self)

    def visitUnarySub(self, node):
        """ Called for unary "-".

        >>> print sourceToString("-100")
        -100
        """

        self.w('-')
        walk(node.expr, self)

    def visitWhile(self, node):
        """ Called for the while statement.

        >>> source = '''
        ... while True:
        ...     some(thing)
        ... else:
        ...     else_(thing)
        ... '''
        >>> print sourceToString(source)
        while True:
            some(thing)
        <BLANKLINE>
        else:
            else_(thing)
        <BLANKLINE>

        """

        self.w('while ')
        walk(node.test, self)
        self.w(':')

        self.indent()
        walk(node.body, self)
        self.dedent()

        if node.else_:
            self.w("else:")
            self.indent()
            walk(node.else_, self)
            self.dedent()

    def visitYield(self, node):
        """ Called for a yield statement.

        >>> print sourceToString("yield y")
        yield y
        """

        self.w('yield ')
        walk(node.value, self)

##    def visitWith(self, node):
##        self.w("with ")
##        walk(node.value, self)
##
    def __str__(self):
        """ Converts the node to a string.  Doing this
        to a Module node will write all of the source code
        for a module. """

        return self.s.getvalue()

    def __getattr__(self, attr):
        """ Called if a node handler is not found. """

        def w(node, attr):
            self.w("Unknown node: %s\n%s" %
                (str(attr), str(node)))

        return lambda node: w(node, attr)


def testSuite():
    import doctest as dt
    import astwriter as moduleToTest

    return dt.DocTestSuite(moduleToTest,
        optionflags = dt.IGNORE_EXCEPTION_DETAIL |
            dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)

if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())
