# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Command extensions for `SetupTools` setup scripts. """

import sys, os, os.path
from glob import glob
from cStringIO import StringIO
from datetime import datetime
from setuptools import *
from distutils.errors import *

from misclibs.functional import *
from misclibs.util import *
from rstlib.astdoc import documentPackages, setDocStringFilter
from traclib import roles, wiki_filter

    
__all__ = ["build_docs", "rst2html", "build_readme", 
    "build_license"]

class rstdocs(Command):
    """ generate python documentation from module sources """

    description = __doc__.strip()
    user_options = [
        ("path=", "p", "Path to docs directory"),
        ("title=", None, "The title of the Package index"),
        ("trac=", None, "The Trac project URL"),
        ("no-status", None, "Supress status messages"),
        ] 
    boolean_options = ["no-status"]
    command_consumes_arguments = True
    
    def initialize_options (self):
        name = os.path.split(os.getcwd())[1]
        self.path = "docs"
        self.no_status = False
        self.title = "%s Package Index" % name
        self.trac = "http:/yoursite.com/trac/project"
        self.args = []
        
    def finalize_options (self):
        roles.setTracRef(roles.tracRef(self.trac))
        setDocStringFilter(wiki_filter.wikiFilter)
        self.path = os.path.abspath(self.path)
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        elif os.path.isfile(self.path):
            raise errors.DistutilsOptionError(
                "The documentation path '%s' is a file "
                "not a directory" % self.path)
                
    def run (self):
        try:
            import docutils
        except:
            raise errors.DistutilsModuleError(
                "The 'package_docs' command requires the docutils "
                "package.")
        
        docPath = os.path.abspath(self.path)
        if self.no_status:
            output = None
        else:
            output = sys.stderr

        docDirs = list(self.args) + sorted(find_packages())

        if not self.dry_run:
            documentPackages(
                docPath = docPath, 
                packages = docDirs,
                writer_name = "html", 
                output=output, 
                title = self.title,
                )


class rst2html(Command):
    """ generate HTML from reStructured text files """

    description = __doc__.strip()
    user_options = []
    boolean_options = []
    command_consumes_arguments = True
    
    def initialize_options (self):
        self.args = []
        
    def finalize_options (self):
        pass
        
    def run (self):
        from docutils.core import publish_file
        
        for srcPath in self.args:
            dstPath = os.path.splitext(srcPath)[0] + ".html"
            
            print "Generating '%s' from '%s'" % (dstPath, srcPath)
            if not self.dry_run:
                publish_file(
                    source_path = srcPath,
                    destination_path = dstPath,
                    writer_name = "html",
                    )

class build_readme(Command):
    """ generate README files from package metadata """

    description = __doc__.strip()
    user_options = []
    boolean_options = []
    
    def initialize_options (self):
        pass
        
    def finalize_options (self):
        pass
        
    def run (self):
        from docutils.core import publish_file
        
        print "Generating README.txt..."
        if not self.dry_run:
            open("README.txt", "w").write('%s%s' % 
                (self.distribution.get_long_description(),
                INSTALLATION)
                )
        print "Generating README.html" 
        if not self.dry_run:
            publish_file(
                source_path = "README.txt",
                destination_path = "README.html",
                writer_name = "html",
                )

class build_license(Command):
    """ generate LICENSE files from package metadata """

    description = __doc__.strip()
    user_options = [
        ("copyright=", None, "Copyright text"),
        ] 
    boolean_options = []
    
    def initialize_options (self):
        self.copyright = str(datetime.today().year)
        
    def finalize_options (self):
        pass
        
    def run (self):
        from docutils.core import publish_file

        copyright = ("Copyright (c) %s %s (%s), " 
            "All rights reserved." %
            (self.copyright, 
            self.distribution.get_author(),
            self.distribution.get_author_email())
            )
            
        license = self.distribution.get_license()
        name = self.distribution.get_name()
        
        if license in LICENSES:
            license_text = LICENSES.get(license) % name
        else:
            license_text = "\n\nPlease contact copyright " \
                "holder for terms of use."
        
        print "Generating LICENSE.txt..."
        if not self.dry_run:
            open("LICENSE.txt", "w").write('%s%s' % 
                (copyright, license_text)
                )
                
        print "Generating LICENSE.html" 
        if not self.dry_run:
            publish_file(
                source_path = "LICENSE.txt",
                destination_path = "LICENSE.html",
                writer_name = "html",
                )


INSTALLATION = """
To build and install from a source distribution execute
the following as root from the root directory of the 
distribution::

    # python setup.py install

    
The following commands can be useful if you wish to modify
the package::
    
    # python setup.py rstdocs           (Build python documentation from module sources)
    # python setup.py build_readme      (Build README files from package Metadata)
    # python setup.py build_license     (Build LICENSE files from package Metadata)
    # python setup.py rst2html          (Convert reStrunctured text files to html)
    
For additional help execute one of the following commands::
    
    python setup.py --help          (for general help)
    python setup.py --help-commands (for a list of commands)
    python setup.py command --help  (for help on a specific command)
"""

LICENSES = {}
LICENSES["MIT"] = """
The %s package is distributed under the terms of the MIT license:

The MIT License

Copyright (c) <year> <copyright holders>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
