# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" reStructured text interpreted text roles for TracLinks. """

import sys, os, os.path, re

from docutils.parsers.rst import roles
from docutils.parsers.rst import directives
from docutils.utils import new_document
from docutils.core import publish_from_doctree
from docutils import nodes, utils

def tracRef(base):
    def tracRef_(*path, **query):
        path = "/".join(path)
        if query:
            query = "?" + "&".join(
                ["%s=%s" % (k,v) for k,v in query.items()])
        else:
            query = ""
        return ("%s/%s%s" % (base, path, query))
    return tracRef_


_tracRef = None
def getTracRef():
    global _tracRef
    return _tracRef
    
def setTracRef(tracRef):
    """ I assign a global tracRef function. The function
    must accept `*path` and `**query` arguments and 
    return a trac reference from the arguments.
    
    >>> setTracRef(tracRef("http:/path/to/my/trac"))
    >>> print getTracRef()("wiki/MyWikiPage", arg1=1)
    http:/path/to/my/trac/wiki/MyWikiPage?arg1=1
    
    """
    global _tracRef
    _tracRef = tracRef

def tracRefError(rawtext, inliner):
    """ Return a reStructured text inline error if
    the trac environment was not initialized.
    
    """
    msg = inliner.reporter.error(
        "No trac reference function set for: %s" \
        % rawtext)
    prb = inliner.problematic(rawtext, rawtext, msg)
    return [prb], [msg]
    
def tracRoleError(errText, inliner):
    """ Return a reStructured text inline error if
    the trac environment was not initialized.
    """
    
    msg = inliner.reporter.error(errText)
    prb = inliner.problematic(rawtext, rawtext, msg)
    return [prb], [msg]

def trac_role(role, rawtext, text, lineno, inliner,
                       options={}, content=[]):
                           
    """ I implement roles for trac wiki syntax. """
    
    global _tracRef
    if _tracRef is None:
        return tracRefError(rawtext, inliner)
        
    if ' ' in text:
        ref, text = text.split(" ", 1)
    else:
        ref = text
        
    if role == "log":
        refs = ref.split(":", 1)
        if len(refs) != 2:
            return roleError(
                "Invalid  log reference: %s" % ref,
                inliner)
        rev, stop_rev = refs
        ref = _tracRef(role, rev=rev, stop_rev=stop_rev)
        
    elif role == 'browser' and '#' in text:
        path, rev = ref.split("#", 1)
        ref = _tracRef(role, path, rev=rev)
        
    else:
        ref = _tracRef(role, ref)
    
    return [nodes.reference(rawtext, utils.unescape(text), 
        refuri=ref)], []
        
roles.register_local_role('wiki', trac_role)
roles.register_local_role('ticket', trac_role)
roles.register_local_role('report', trac_role)
roles.register_local_role('changeset', trac_role)
roles.register_local_role('log', trac_role)
roles.register_local_role('milestone', trac_role)
roles.register_local_role('browser', trac_role)
roles.register_local_role('pydoc', trac_role)
roles.register_local_role('apidoc', trac_role)

def arguments(name, args, options, content, lineno,
             content_offset, block_text, state, state_machine):
    text = '\n'.join(content)
    if not text:
        error = state_machine.reporter.error(
            'The "%s" directive is empty; content required.' % name,
            nodes.literal_block(block_text, block_text), line=lineno)
        return [error]
    node = nodes.compound(text)
    node['classes'] += options.get('class', [])
    state.nested_parse(content, content_offset, node)
    if args:
        note = " (%s):" % args[0]
    else:
        note = ":"
    return  nodes.block_quote("",
                nodes.paragraph("", 
                    "Arguments %s" % note, 
                    node))
                    
arguments.arguments = (0, 1 , 1)
arguments.options = {'class': directives.class_option}
arguments.content = 1
directives.register_directive("arguments", arguments)

def testSuite():
    """ Create a unit test suite. """
    
    import doctest as dt
    import roles as moduleToTest
    
    return dt.DocTestSuite(moduleToTest, 
        optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
            dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)
    
if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())
