# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Unit Tests for the wiki_filter module.  """

import sys, os, os.path, re
path =  os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir))
                
if path not in sys.path:
    sys.path.append(path)
    
import wiki_filter
from wiki_filter import wikiFilter

def _testFilter(docstring):
    """ Filter a docstring for wiki links. """
    
    lines = []
    for line in docstring.splitlines():
        if not line.strip():
            break
        lines.append(line)
    return wikiFilter("\n".join(lines))
   
def testWikiLinks():
    """ This is a wiki link: `aWikiPage This is displayed`. 
    
    >>> print _testFilter(testWikiLinks.__doc__)
    This is a wiki link: :wiki:`aWikiPage This is displayed`.

    """

def testWikiEscapes():
    """ This is not a wiki link: !`aWikiPage This is displayed`.
    
    >>> print _testFilter(testWikiEscapes.__doc__)
    This is not a wiki link: \`aWikiPage This is displayed\`.

    """

    
def testTicketLinks():
    """ This is a trac ticket #123. 
    
    >>> print _testFilter(testTicketLinks.__doc__)
    This is a trac ticket :ticket:`123 #123`.

    """
    

def testTicketEscapes():
    """ This is not a trac ticket !#123. 
    
    >>> print _testFilter(testTicketEscapes.__doc__)
    This is not a trac ticket #123.
    
    """
    
def testReportLinks():
    """ This is a trac report {123}. 
    
    >>> print _testFilter(testReportLinks.__doc__)
    This is a trac report :report:`123 {123}`.

    """

def testReportEscapes():
    """ This is not a trac report !{123}. 
    
    >>> print _testFilter(testReportEscapes.__doc__)
    This is not a trac report {123}. 

    """
   
def testLogLinks():
    """ This is a trac revision log: r123:456. 
    And this: [123:456]. 
        
    >>> print _testFilter(testLogLinks.__doc__)
    This is a trac revision log: :log:`123:456 r123:456`.
    And this: :log:`123:456 [123:456]`. 
    
    """

def testLogEscapes():
    """ This is not a trac revision log: !r123:456. 
    And not this: ![123:456].  
    
    >>> print _testFilter(testLogEscapes.__doc__)
    This is not a trac revision log: r123:456.
    And not this: [123:456].  
    
    """
    
def testChgSetLinks():
    """ This is a trac changeset: r123. And this: [123]. 
    
    >>> print _testFilter(testChgSetLinks.__doc__)
    This is a trac changeset: :changeset:`123 r123` 
    And this: :changeset:`123 [123]`.
    
    """

def testChgSetEscapes():
    """ This is not a trac changeset: !r123. 
    And not this: ![123]. 
    
    >>> print _testFilter(testChgSetEscapes.__doc__)
    This is not a trac changeset: r123. 
    And not this: [123].
    
    """

def testAll():
    """ Test all,  from examples in wiki_fileter
    
    >>> print wikiFilter(wiki_filter.__doc__)
     Convert :wiki:`TracLinks` into :wiki:`RestructuredText` roles.  The 
    following examples illustrate the use of :wiki:`TracLinks` in 
    Python :wiki:`DocStrings`.
    <BLANKLINE>
        - \`AWikiPage\`: A link to :wiki:`AWikiPage`
        - #1: A link to a ticket :ticket:`1 #1`
        - {1}: A link to a report :report:`1 {1}`
        - r1 or [1]: A link to a change set :changeset:`1 r1` or :changeset:`1 [1]`
        - r1:2 or [1:2]: A link to a revision log :log:`1:2 r1:2` or :log:`1:2 [1:2]`
        - ``:browser:`trunk/some/file.py source repository```:
          A link to the :browser:`trunk/some/file.py source repository`.
    <BLANKLINE>
    The preceeding can be escaped by prepending them with a '!' (bang).
    Note that change set and log links do not currently work for
    the bzr repository backend because they do not use numeric
    revisions.
    <BLANKLINE>
    """
    
def testSuite():
    import doctest as dt
    
    return dt.DocTestSuite(
        optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
            dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)
    
if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())

