# -*- coding: utf-8 -*-
#
# Copyright Shawn L. Church (c) 2008, All rights reserved.
# For terms of use and distribution see the file LICENSE.txt
# located in the root directory of the distribution.

""" Convert `TracLinks` into `RestructuredText` roles.  The 
following examples illustrate the use of `TracLinks` in 
Python `DocStrings`.
    
    - !`AWikiPage`: A link to `AWikiPage`
    - #1: A link to a ticket #1
    - !{1}: A link to a report {1}
    - !r1 or ![1]: A link to a change set r1. or [1]
    - !r1:2 or ![1:2]: A link to a revision log r1:2 or [1:2]
    - ``:browser:`trunk/some/file.py source repository```:
      A link to the :browser:`trunk/some/file.py source repository`.

The preceeding can be escaped by prepending them with a '!' (bang).
Note that change set and log links do not currently work for
the bzr repository backend because they do not use numeric
revisions.
"""

import re
from itertools import chain

patterns = [
    ################################################
    # These patterns replace trac wiki syntax with
    # restructured text custom roles.
    ################################################

    # Wiki Links `A Wiki Page`:
    (r"(\s|\A)`([^`]+)`", r"\1:wiki:`\2`"),
    
    # TicketLinks #123:
    (r"(\s|\A)#([\d]+)([^:\d])", r"\1:ticket:`\2 #\2`\3"),
    
    # Report Links {123}:
    (r"(\s|\A)\{(\d+)\}", r"\1:report:`\2 {\2}`"),
    
    # Revision log r123:456 or [123:456]:
    (r"(\s|\A)r(\d+:\d+)", r"\1:log:`\2 r\2`"),
    (r"(\s|\A)\[(\d+:\d+)\]", r"\1:log:`\2 [\2]`"),
    
    # Changesets r123 or [123]:
    (r"(\s|\A)r(\d+)[^\d:]", r"\1:changeset:`\2 r\2`"),
    (r"(\s|\A)\[(\d+)\]", r"\1:changeset:`\2 [\2]`"),
    
    ]

escapes = [
    ################################################
    # These patterns escape the patterns above
    # (i.e. !pattern is replaced with pattern)
    ################################################

    # Wiki Links:
    (r"(\s|\A)!`([^`]+)`", r"\1\\`\2\\`"),
    
    # Ticket Links:
    (r"!#(\d+)([^:\d])", r"#\1\2"),
    
    # Report Links:
    (r"!(\{\d+\})", r"\1"),
    
    # Revision logs:
    (r"!(r\d+:\d+)", r"\1"),
    (r"!(\[\d+:\d+\])", r"\1"),
    
    # Changesets:
    (r"!(r\d+)([^\d:])", r"\1\2"),
    (r"!(\[\d+\])", r"\1"),
    ]

# Compile the regular expressions.
patterns = [(re.compile(search), replace)
            for search, replace in patterns] 
escapes = [(re.compile(escape), replace)
            for escape, replace in escapes] 
            
def wikiFilter(string):
    """ Convert Trac Wiki syntax to reStructured Text
    interpreted text roles.
        
    """
    
    for search, replace in chain(patterns, escapes):
        string = search.sub(replace, string)
    return string

def testSuite():
    import doctest as dt
    import test_wiki_filter as moduleToTest
    
    return dt.DocTestSuite(moduleToTest, 
        optionflags = dt.IGNORE_EXCEPTION_DETAIL | 
            dt.ELLIPSIS | dt.NORMALIZE_WHITESPACE)
    
if __name__ == "__main__":
    import unittest
    unittest.TextTestRunner(verbosity=2).run(testSuite())